import numpy as np

from paretoset.algorithms_numpy import paretoset_efficient, pareto_rank_naive
from paretoset.utils import user_has_package, validate_inputs


import pandas as pd

if user_has_package("numba"):
    from paretoset.algorithms_numba import BNL


def paretoset(costs, sense=None, distinct=True, use_numba=True):
    """Return boolean mask indicating the Pareto set of (non-NaN) numerical data.

    The input data in `costs` can be either a pandas DataFrame or a NumPy ndarray
    of shape (observations, objectives). The user is responsible for dealing with
    NaN values *before* sending data to this function. Only numerical data is
    allowed, with the exception of `diff` (different) columns.

    Parameters
    ----------
    costs : np.ndarray or pd.DataFrame
        Array or DataFrame of shape (observations, objectives).
    sense : list
        List with strings for each column (objective). The value `min` (default)
        indicates minimization, `max` indicates maximization and `diff` indicates
        different values. Using `diff` is equivalent to a group-by operation
        over the columns marked with `diff`. If None, minimization is assumed.
    distinct : bool
        How to treat duplicate rows. If `True`, only the first duplicate is returned.
        If `False`, every identical observation is returned instead.
    use_numba : bool
        If True, numba will be used if it is installed by the user.

    Returns
    -------
    mask : np.ndarray
        Boolean mask with `True` for observations in the Pareto set.

    Examples
    --------
    >>> from paretoset import paretoset
    >>> import numpy as np
    >>> costs = np.array([[2, 0], [1, 1], [0, 2], [3, 3]])
    >>> paretoset(costs)
    array([ True,  True,  True, False])
    >>> paretoset(costs, sense=["min", "max"])
    array([False, False,  True,  True])

    The `distinct` parameter:

    >>> paretoset([0, 0], distinct=True)
    array([ True, False])
    >>> paretoset([0, 0], distinct=False)
    array([ True,  True])
    """

    if user_has_package("numba") and use_numba:
        paretoset_algorithm = BNL
    else:
        paretoset_algorithm = paretoset_efficient

    costs, sense = validate_inputs(costs=costs, sense=sense)
    assert isinstance(sense, list)

    n_costs, n_objectives = costs.shape

    diff_cols = [i for i in range(n_objectives) if sense[i] == "diff"]
    max_cols = [i for i in range(n_objectives) if sense[i] == "max"]
    min_cols = [i for i in range(n_objectives) if sense[i] == "min"]

    # Check data types (MIN and MAX must be numerical)
    message = "Data must be numerical. Please convert it. Data has type: {}"

    if isinstance(costs, pd.DataFrame):
        data_types = [costs.dtypes.values[i] for i in (max_cols + min_cols)]
        if any(d == np.dtype("O") for d in data_types):
            raise TypeError(message.format(data_types))

    else:
        if costs.dtype == np.dtype("O"):
            raise TypeError(message.format(costs.dtype))

    # CASE 1: THE ONLY SENSE IS MINIMIZATION
    # ---------------------------------------
    if all(s == "min" for s in sense):
        if isinstance(costs, pd.DataFrame):
            costs = costs.to_numpy(copy=True)
        return paretoset_algorithm(costs, distinct=distinct)

    n_costs, n_objectives = costs.shape

    if not diff_cols:
        # Its an array
        if not isinstance(costs, np.ndarray):
            costs = costs.to_numpy(copy=True)
        for col in max_cols:
            costs[:, col] = -costs[:, col]

        return paretoset_algorithm(costs, distinct=distinct)

    if isinstance(costs, pd.DataFrame):
        df = costs.copy()  # Copy to avoid mutating inputs
        df.columns = np.arange(n_objectives)
    else:
        df = pd.DataFrame(costs)

    assert isinstance(df, pd.DataFrame)
    assert np.all(df.columns == np.arange(n_objectives))

    # If `object` columns are present and they can be converted, do it.
    for col in max_cols + min_cols:
        df[col] = pd.to_numeric(df[col], errors="coerce")

    is_efficient = np.zeros(n_costs, dtype=np.bool_)

    # Create the groupby object
    # We could've implemented our own groupby, but choose to use pandas since
    # it's likely better than what we can come up with on our own.
    groupby = df.groupby(diff_cols)

    # Iteration through the groups
    for key, data in groupby:

        # Get the relevant data for the group and compute the efficient points
        relevant_data = data[max_cols + min_cols].to_numpy(copy=True)
        efficient_mask = paretoset_algorithm(relevant_data.copy(), distinct=distinct)

        # The `pd.DataFrame.groupby.indices` dict holds the row indices of the group
        data_mask = groupby.indices[key]
        is_efficient[data_mask] = efficient_mask

    return is_efficient


def paretorank(costs, sense=None, distinct=True, use_numba=True):
    """Return integer array with Pareto ranks of (non-NaN) numerical data.

    Observations in the Pareto set are assigned rank 1. After removing the Pareto
    set, the Pareto set of the remaining data is assigned rank 2, and so forth.

    The input data in `costs` can be either a pandas DataFrame or a NumPy ndarray
    of shape (observations, objectives). The user is responsible for dealing with
    NaN values *before* sending data to this function. Only numerical data is
    allowed, with the exception of `diff` (different) columns.

    Parameters
    ----------
    costs : np.ndarray or pd.DataFrame
        Array or DataFrame of shape (observations, objectives).
    sense : list
        List with strings for each column (objective). The value `min` (default)
        indicates minimization, `max` indicates maximization and `diff` indicates
        different values. Using `diff` is equivalent to a group-by operation
        over the columns marked with `diff`. If None, minimization is assumed.
    distinct : bool
        How to treat duplicate rows. If `True`, only the first duplicate is returned.
        If `False`, every identical observation is returned instead.
    use_numba : bool
        If True, numba will be used if it is installed by the user.

    Returns
    -------
    ranks : np.ndarray
        Integer array with Pareto ranks of the observations.

    Examples
    --------
    >>> from paretoset import paretoset
    >>> import numpy as np
    >>> costs = np.array([[2, 0], [1, 1], [0, 2], [3, 3]])
    >>> paretorank(costs)
    array([1, 1, 1, 2])
    >>> paretorank(costs, sense=["min", "max"])
    array([3, 2, 1, 1])

    The `distinct` parameter:

    >>> paretorank([0, 0], distinct=True)
    array([1, 2])
    >>> paretorank([0, 0], distinct=False)
    array([1, 1])
    """

    if user_has_package("numba") and use_numba:
        paretorank_algorithm = pareto_rank_naive
    else:
        paretorank_algorithm = pareto_rank_naive

    costs, sense = validate_inputs(costs=costs, sense=sense)
    assert isinstance(sense, list)

    n_costs, n_objectives = costs.shape

    diff_cols = [i for i in range(n_objectives) if sense[i] == "diff"]
    max_cols = [i for i in range(n_objectives) if sense[i] == "max"]
    min_cols = [i for i in range(n_objectives) if sense[i] == "min"]

    # Check data types (MIN and MAX must be numerical)
    message = "Data must be numerical. Please convert it. Data has type: {}"

    if isinstance(costs, pd.DataFrame):
        data_types = [costs.dtypes.values[i] for i in (max_cols + min_cols)]
        if any(d == np.dtype("O") for d in data_types):
            raise TypeError(message.format(data_types))

    else:
        if costs.dtype == np.dtype("O"):
            raise TypeError(message.format(costs.dtype))

    # CASE 1: THE ONLY SENSE IS MINIMIZATION
    # ---------------------------------------
    if all(s == "min" for s in sense):
        if isinstance(costs, pd.DataFrame):
            costs = costs.to_numpy(copy=True)
        return paretorank_algorithm(costs, distinct=distinct, use_numba=use_numba)

    n_costs, n_objectives = costs.shape

    if not diff_cols:
        # Its an array
        if not isinstance(costs, np.ndarray):
            costs = costs.to_numpy(copy=True)
        for col in max_cols:
            costs[:, col] = -costs[:, col]

        return paretorank_algorithm(costs, distinct=distinct, use_numba=use_numba)

    if isinstance(costs, pd.DataFrame):
        df = costs.copy()  # Copy to avoid mutating inputs
        df.columns = np.arange(n_objectives)
    else:
        df = pd.DataFrame(costs)

    assert isinstance(df, pd.DataFrame)
    assert np.all(df.columns == np.arange(n_objectives))

    # If `object` columns are present and they can be converted, do it.
    for col in max_cols + min_cols:
        df[col] = pd.to_numeric(df[col], errors="coerce")

    all_ranks = np.zeros(n_costs, dtype=np.int_)

    # Create the groupby object
    # We could've implemented our own groupby, but choose to use pandas since
    # it's likely better than what we can come up with on our own.
    groupby = df.groupby(diff_cols)

    # Iteration through the groups
    for key, data in groupby:

        # Get the relevant data for the group and compute the efficient points
        relevant_data = data[max_cols + min_cols].to_numpy(copy=True)
        ranks = paretorank_algorithm(relevant_data.copy(), distinct=distinct, use_numba=use_numba)

        # The `pd.DataFrame.groupby.indices` dict holds the row indices of the group
        data_mask = groupby.indices[key]
        all_ranks[data_mask] = ranks

    return all_ranks


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "--maxfail=5", "--cache-clear", "--color", "yes", ""])



import importlib
import numpy as np
import collections.abc


def user_has_package(package_name):
    """Check if the user has `numba` installed."""
    try:
        importlib.import_module(package_name, package=None)
        return True
    except ModuleNotFoundError:
        return False


def validate_inputs(costs, sense=None):
    """Sanitize user inputs for the `paretoset` function.

    Examples
    --------
    >>> costs, sense = validate_inputs([1, 2, 3])
    >>> costs
    array([[1],
           [2],
           [3]])

    """

    # The input is an np.ndarray
    if isinstance(costs, np.ndarray):
        if costs.ndim == 1:
            return validate_inputs(costs.copy().reshape(-1, 1), sense=sense)
        if costs.ndim != 2:
            raise ValueError("`costs` must have shape (observations, objectives).")

        # It's a 2D ndarray -> copy it
        costs = costs.copy()

    elif user_has_package("pandas"):
        import pandas as pd

        if not isinstance(costs, pd.DataFrame):
            return validate_inputs(np.asarray(costs), sense=sense)
    else:
        return validate_inputs(np.asarray(costs), sense=sense)

    # if (not (isinstance(costs, np.ndarray) and costs.ndim == 2):
    #    raise TypeError("`costs` must be a NumPy array with 2 dimensions or pandas DataFrame.")

    n_costs, n_objectives = costs.shape

    if sense is None:
        return costs, ["min"] * n_objectives
    else:
        sense = list(sense)

    if not isinstance(sense, collections.abc.Sequence):
        raise TypeError("`sense` parameter must be a sequence (e.g. list).")

    if not len(sense) == n_objectives:
        raise ValueError("Length of `sense` must match second dimensions (i.e. columns).")

    # Convert functions "min" and "max" to their names
    sense = [s.__name__ if callable(s) else s for s in sense]
    if not all(isinstance(s, str) for s in sense):
        raise TypeError("`sense` parameter must be a sequence of strings.")
    sense = [s.lower() for s in sense]

    sense_map = {"min": "min", "minimum": "min", "max": "max", "maximum": "max", "diff": "diff", "different": "diff"}

    sense = [sense_map.get(s) for s in sense]

    # Verify that the strings are of correct format
    valid = ["min", "max", "diff"]
    if not all(s in valid for s in sense):
        raise TypeError("`sense` must be one of: {}".format(valid))

    return costs, sense


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "--maxfail=5", "--cache-clear", "--color", "yes", ""])



import numpy as np
import paretoset.user_interface  # Import like this to avoid circular import issues


def paretoset_naive(costs, distinct=True):
    """Naive implementation.

    Parameters
    ----------
    costs : (np.ndarray) Array of shape (n_costs, n_objectives).

    Returns
    -------
    mask : (np.ndarray) Boolean array indicating the paretoset.

    """
    n_costs, n_objectives = costs.shape

    # Assume all points/costs are inefficient
    is_efficient = np.zeros(n_costs, dtype=np.bool_)

    for i in range(n_costs):
        this_cost = costs[i, :]

        # Here `NOT(ANY(a_i > b_i))` is equal to ALL(a_i <= b_i), but faster.
        at_least_as_good = np.logical_not(np.any(costs > this_cost, axis=1))
        any_better = np.any(costs < this_cost, axis=1)

        dominated_by = np.logical_and(at_least_as_good, any_better)

        # If we're looking for distinct points and it's already in the
        # pareto set, disregard this value.
        if distinct and np.any(is_efficient):
            if np.any(np.all(costs[is_efficient] == this_cost, axis=1)):
                continue

        if not (np.any(dominated_by[:i]) or np.any(dominated_by[i + 1 :])):
            is_efficient[i] = True

    return is_efficient


def paretoset_efficient(costs, distinct=True):
    """An efficient vectorized algorhtm.

    This algorithm was given by Peter in this answer on Stack Overflow:
    - https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python
    """
    costs = costs.copy()  # The algorithm mutates the `costs` variable, so we take a copy
    n_costs, n_objectives = costs.shape

    is_efficient = np.arange(n_costs)

    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):

        this_cost = costs[next_point_index]

        # Two points `a` and `b` are *incomparable* if neither dom(a, b) nor dom(b, a).
        # Points that are incomparable to `this_cost`, or dominate `this_cost`.
        # In 2D, these points are below or to the left of `this_cost`.
        current_efficient_points = np.any(costs < this_cost, axis=1)

        # If we're not looking for distinct, keep points equal to this cost
        if not distinct:
            no_smaller = np.logical_not(current_efficient_points)
            equal_to_this_cost = np.all(costs[no_smaller] == this_cost, axis=1)
            current_efficient_points[no_smaller] = np.logical_or(
                current_efficient_points[no_smaller], equal_to_this_cost
            )

        # Any point is incomparable to itself, so keep this point
        current_efficient_points[next_point_index] = True

        # Remove dominated points
        is_efficient = is_efficient[current_efficient_points]
        costs = costs[current_efficient_points]

        # Re-adjust the index
        next_point_index = np.sum(current_efficient_points[:next_point_index]) + 1

    # Create a boolean mask from indices and return it
    is_efficient_mask = np.zeros(n_costs, dtype=np.bool_)
    is_efficient_mask[is_efficient] = True
    return is_efficient_mask


def pareto_rank_naive(costs, distinct=True, use_numba=True):
    """Naive implementation of Pareto ranks."""

    n_costs, n_objectives = costs.shape
    ranks = np.zeros(n_costs, dtype=np.int_)
    remaining = np.ones(n_costs, dtype=np.bool_)
    processed = np.zeros(n_costs, dtype=np.bool_)

    current_rank = 1
    while np.sum(remaining) > 0:

        # Mark the costs that have rank `i`
        frontier_mask = paretoset.user_interface.paretoset(costs[remaining], distinct=distinct, use_numba=use_numba)

        # Processed costs in this iteration (not processed already, and in the frontier)
        processed[np.logical_not(processed)] = frontier_mask

        # Set the rank of the processed costs that are still remaining (remaining not updated yet)
        ranks[np.logical_and(processed, remaining)] = current_rank

        # Update the remaining values
        remaining[remaining] = np.logical_not(frontier_mask)
        current_rank += 1

    return ranks


def crowding_distance(costs):
    """Compute the crowding distance of (non-NaN) numerical data.

    The input data in `costs` must be a NumPy ndarray of shape
    (observations, objectives). The user is responsible for dealing with NaN
    values, duplicate rows and constant columns *before* calling this function.

    Parameters
    ----------
    costs : np.ndarray
        Numerical data of shape (observations, objectives).

    Returns
    -------
    distances : np.ndarray
        The crowding distance of each observation (row).

    Examples
    --------
    >>> import numpy as np
    >>> costs = np.array([1, 3, 5, 9, 11]).reshape(-1, 1)
    >>> crowding_distance(costs)
    array([inf, 0.4, 0.6, 0.6, inf])
    """
    msg = "The `costs` argument must be np.ndarray of shape (observations, objectives)"
    if not isinstance(costs, np.ndarray):
        raise ValueError(msg)
    if not costs.ndim == 2:
        raise ValueError(msg)

    # =============================================================================
    # SETUP ARRAYS USED FOR COMPUTATION
    # =============================================================================

    # Get the shape of the inputs
    n_costs, n_objectives = costs.shape
    if n_costs <= 2:
        return np.ones(n_costs) * np.inf

    # Prepare the distance matrix
    distances = np.zeros_like(costs, dtype=np.float)  # Must use floats to allow np.inf

    # Used several times below, so pre-compute it here
    arange_objectives = np.arange(n_objectives)

    # Used to inverse the sort
    sorted_inds_reversed = np.zeros(costs.shape, dtype=np.int)

    # =============================================================================
    # PERFORM THE COMPUTATION
    # =============================================================================

    # Sort every column individually
    sorted_inds = np.argsort(costs, axis=0)
    sorted_matrix = np.take_along_axis(costs, sorted_inds, axis=0)

    # Compute the diff of a_i as a_{i+1} - a_{i-1}
    diffs = sorted_matrix[2:, :] - sorted_matrix[:-2, :]

    # Compute max minus min for each column, using the already sorted matrix
    max_minus_min = sorted_matrix[-1, :] - sorted_matrix[0, :]

    # Bottom and top rows are set to infinity - these points have 1 neighbor
    infinity_row = (np.ones(n_objectives) * np.inf).reshape(1, -1)

    # Create the full matrix of differences
    diffs = np.vstack((infinity_row, diffs, infinity_row))
    assert diffs.shape == costs.shape

    # Prepare a matrix of reverse-sorted indices
    index = sorted_inds, arange_objectives

    # Equivalent to: np.outer(np.arange(n_costs), np.ones(n_objectives))
    sorted_inds_reversed[index] = np.tile(np.arange(n_costs), (n_objectives, 1)).T

    # Distances (diffs) in original data order
    distances = diffs[sorted_inds_reversed, arange_objectives]

    # Divide each column by MAX minus MIN to normalize
    distances = np.divide(distances, max_minus_min, out=distances)

    # Sum over each row, divide by number of columns and return
    return distances.sum(axis=1) / n_objectives


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "--color", "yes"])

if __name__ == "__main__":

    pass



import numpy as np
import numba


@numba.jit(nopython=True)
def any_jitted(costs, cost):
    """Check if any are smaller over axis 1."""

    rows, cols = costs.shape
    ans = np.zeros(shape=rows, dtype=np.bool_)

    for i in range(rows):
        for j in range(cols):
            if costs[i, j] < cost[j]:
                ans[i] = True
                break  # Break out early here
    return ans


@numba.jit(nopython=True)
def all_jitted(costs, cost):
    """Check if all are equal over axis 1."""

    rows, cols = costs.shape
    ans = np.ones(shape=rows, dtype=np.bool_)

    for i in range(rows):
        for j in range(cols):
            if costs[i, j] != cost[j]:
                ans[i] = False
                break  # Break out early here
    return ans


@numba.jit(nopython=True)
def any_equal_jitted(costs, cost):
    """Check if any are equal over axis 1."""

    rows, cols = costs.shape

    for i in range(rows):
        equal = True  # Assume equality
        for j in range(cols):
            if costs[i, j] != cost[j]:
                equal = False
                break  # Break out early here

        # This row in `costs` was equal to `cost`, return immediately
        if equal:
            return True

    return False


@numba.jit(nopython=True)
def paretoset_jit(costs, distinct=True):
    """
    Find the pareto-efficient points
    :param costs: An (n_points, n_costs) array
    :param return_mask: True to return a mask
    :return: An array of indices of pareto-efficient points.
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.
    """
    # https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python

    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]

    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):

        this_cost = costs[next_point_index]

        # Keep any point with a lower cost
        current_efficient_points = any_jitted(costs, this_cost)
        # np.any(costs < costs[next_point_index], axis=1)
        current_efficient_points[next_point_index] = True  # And keep self

        # If we're not looking for distinct, keep points equal to this cost
        if not distinct:
            no_smaller = np.logical_not(current_efficient_points)
            equal_to_this_cost = all_jitted(costs[no_smaller], this_cost)
            current_efficient_points[no_smaller] = np.logical_or(
                current_efficient_points[no_smaller], equal_to_this_cost
            )

        # Remove dominated points
        is_efficient = is_efficient[current_efficient_points]
        costs = costs[current_efficient_points]

        # Re-adjust the index
        next_point_index = np.sum(current_efficient_points[:next_point_index]) + 1

    is_efficient_mask = np.zeros(shape=n_points, dtype=np.bool_)
    is_efficient_mask[is_efficient] = True
    return is_efficient_mask


@numba.jit(nopython=True, fastmath=True)
def dominates(a, b, length):
    """Does a dominate b?"""
    better = False
    for i in range(length):
        a_i, b_i = a[i], b[i]

        # Worse in one dimension -> does not domiate
        # This is faster than computing `at least as good` in every dimension
        if a_i > b_i:
            return False

        # Better in at least one dimension
        if a_i < b_i:
            better = True
    return better


@numba.jit(nopython=True, fastmath=True)
def window_dominates_cost(window, cost, window_rows, window_cols):
    for i in range(window_rows):
        if dominates(window[i], cost, window_cols):
            return i
    return -1


@numba.jit(nopython=True, fastmath=True)
def cost_dominates_window(window, cost, window_rows, window_cols):
    dominated_inds = []
    for i in range(window_rows):
        if dominates(cost, window[i], window_cols):
            dominated_inds.append(i)
    return dominated_inds


@numba.jit(nopython=True)
def BNL(costs, distinct=True):
    """
    Block nested loops algorithm.
    """

    is_efficient = np.arange(costs.shape[0])
    n_costs, n_objectives = costs.shape
    num_efficient = 1  # Always put the first row in the window

    window_changed = True

    for i in range(1, n_costs):  # Skip the first row, since it's in the window

        # Get the cost for this row
        this_cost = costs[i]

        # If the window indices changed in the last iteration, get window again
        if window_changed:
            window = costs[is_efficient[:num_efficient]]
            window_rows, window_cols = window.shape
            window_changed = False

        # CASE 1 : DOES ANYTHING IN THE WINDOW DOMINATE THIS COST?
        # --------------------------------------------------------

        dom_index = window_dominates_cost(window, this_cost, window_rows, window_cols)
        # `dom_index` is the index of the first window element that dominates
        # the cost. If no window elements dominate the cost, -1 is returned.
        if dom_index >= 0:
            continue  # Window dominates cost, move on.

        # CASE 2 : DOES THIS COST DOMINATE ANYTHING IN THE WINDOW?
        # --------------------------------------------------------

        # Check if anything in the window is dominated by the point in question
        dominated_inds_window = cost_dominates_window(window, this_cost, window_rows, window_cols)
        # A point in the window is dominated, remove it
        if len(dominated_inds_window) != 0:

            # Get the original indices to remove
            to_removes = [is_efficient[k] for k in dominated_inds_window]
            for to_remove in to_removes:

                # Original indices of elements in the window
                for j, efficient in enumerate(is_efficient):

                    # Found a match, remove it
                    if efficient == to_remove:
                        # Move one to the left and decrement
                        is_efficient[j:num_efficient] = is_efficient[j + 1 : num_efficient + 1]
                        num_efficient -= 1
                        break  # Break out here

        # CASE 3 : ADD THE NEW COST TO THE WINDOW
        # ---------------------------------------

        # Add to window in all cases, except if `distinct` and it's already in the window
        if (not distinct) or (not any_equal_jitted(window, this_cost)):

            # Insert the index value of the point in the last position
            is_efficient[num_efficient] = i
            # Increment the number of efficient points
            num_efficient += 1
            window_changed = True

    bools = np.zeros(costs.shape[0], dtype=np.bool_)
    bools[is_efficient[:num_efficient]] = 1
    return bools


# @numba.jit(nopython=True)
def pareto_rank_NSGA2(costs):
    """Algorithm from the NSGA-II paper."""
    n_costs, n_objectives = costs.shape

    dominated_by = [set() for i in range(n_costs)]

    domination_counter = np.zeros(n_costs, dtype=np.int_)
    ranks = np.zeros(n_costs, dtype=np.int_)

    domination_counter2 = np.zeros(n_costs, dtype=np.int_)
    dominated_by2 = [set() for i in range(n_costs)]

    frontier = set()
    for i in range(n_costs):
        this_cost = costs[i, :]

        dominated_by_i = np.all(this_cost >= costs, axis=1) & np.any(this_cost > costs, axis=1)
        dominated_by2[i].update(set([i for i in range(n_costs) if dominated_by_i[i]]))

        dominates_i = np.all(costs <= this_cost, axis=1) & np.any(costs < this_cost, axis=1)
        domination_counter2[i] += np.sum(dominates_i)

        for j in range(n_costs):
            other_cost = costs[j, :]

            if dominates(this_cost, other_cost, n_objectives):
                # Add `other_cost` to the set of solutions dominated by `this_cost`
                dominated_set = dominated_by[i]
                dominated_set.add(j)
            elif dominates(other_cost, this_cost, n_objectives):
                # Increment domination counter of `this_cost`
                domination_counter[i] += 1

        assert (domination_counter == domination_counter2).all()
        assert [i == j for i, j in zip(dominated_by, dominated_by2)]
        if domination_counter[i] == 0:
            ranks[i] = 1
            frontier.add(i)

    rank = 2
    while frontier:
        new_frontier = set()
        # Get all points
        for frontier_i in frontier:
            for dominated_j in dominated_by[frontier_i]:
                domination_counter[dominated_j] -= 1
                if domination_counter[dominated_j] == 0:
                    ranks[dominated_j] = rank
                    new_frontier.add(dominated_j)
        rank += 1
        frontier = new_frontier

    return ranks


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "--maxfail=5", "--color", "yes"])
